﻿//
// MainPage.xaml.cpp
// Implementation of the MainPage class.
//

#include "pch.h"
#include "MainPage.xaml.h"

using namespace DeeJay;

using namespace Concurrency;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::ApplicationModel;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;
using namespace Windows::Storage;
using namespace Windows::Storage::FileProperties;
using namespace Windows::Storage::Pickers;
using namespace Windows::Storage::Streams;
using namespace Windows::UI::Core;
using namespace Windows::UI::Popups;
using namespace Windows::UI::ViewManagement;
using namespace Windows::UI::Xaml;
using namespace Windows::UI::Xaml::Controls;
using namespace Windows::UI::Xaml::Controls::Primitives;
using namespace Windows::UI::Xaml::Data;
using namespace Windows::UI::Xaml::Input;
using namespace Windows::UI::Xaml::Media;
using namespace Windows::UI::Xaml::Media::Imaging;
using namespace Windows::UI::Xaml::Navigation;

MainPage::MainPage()
{
	InitializeComponent();

    recordEllipse->ManipulationMode = ManipulationModes::Rotate | ManipulationModes::RotateInertia;

    // Start up the Media Foundation
    HRESULT hresult = MFStartup(MF_VERSION);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFStartup failure");

    // Create an IXAudio2 object
    hresult = XAudio2Create(&pXAudio2);

    if (FAILED(hresult))
        ref new COMException(hresult, "XAudio2Create failure");

    // Create a mastering voice
    hresult = pXAudio2->CreateMasteringVoice(&pMasteringVoice); 

    if (FAILED(hresult))
        ref new COMException(hresult, "IXAudio2->CreateMasteringVoice failure");

    // Initialize the volume
    hresult = pMasteringVoice->SetVolume(0);

    if (FAILED(hresult))
        ref new COMException(hresult, "IXAudio2MasteringVoice->SetVolume failure");

    SizeChanged += ref new SizeChangedEventHandler(this, &MainPage::OnSizeChanged);
    CompositionTarget::Rendering += ref new EventHandler<Object^>(this, &MainPage::OnCompositionTargetRendering);
}

void MainPage::OnSizeChanged(Object^ sender, SizeChangedEventArgs^ args)
{
    // Make sure we can't try to load files in snapped view
    openButton->IsEnabled = ApplicationView::Value != ApplicationViewState::Snapped;
}

void MainPage::OnCompositionTargetRendering(Object^ sender, Object^ args)
{
    static double lastSecondsElapsed;

    if (pAudioFilePlayer != nullptr)
    {
        double secondsElapsed = pAudioFilePlayer->GetSecondsElapsed();

        if (secondsElapsed != lastSecondsElapsed)
        {
            // 33-1/3 RPM == 200 degrees/second
            recordRotate->Angle = 200 * secondsElapsed;

            // Move the needle to the current location
            needleTransform->TranslateX = MAX_GROOVE_RADIUS -
                secondsElapsed * (MAX_GROOVE_RADIUS - MIN_GROOVE_RADIUS) / maxDurationInSeconds;

            // Wiggle the needle a bit
            needleTransform->Rotation = rand() % 3 - 1;
            lastSecondsElapsed = secondsElapsed;
        }
    }
}

// Appliation bar control handlers
void MainPage::OnSpeedRatioSliderValueChanged(Object^ sender, RangeBaseValueChangedEventArgs^ args)
{
    if (pAudioFilePlayer != nullptr)
    {
        pAudioFilePlayer->SetSpeedRatio((float)args->NewValue);
    }
}

void MainPage::OnVolumeSliderValueChanged(Object^ sender, RangeBaseValueChangedEventArgs^ args)
{
    HRESULT hresult = pMasteringVoice->SetVolume(float(args->NewValue / 100));

    if (FAILED(hresult))
        ref new COMException(hresult, "IXAudio2MasteringVoice->SetVolume failure");
}

void MainPage::OnOpenButtonClick(Object^ sender, RoutedEventArgs^ args)
{
    FileOpenPicker^ fileOpenPicker = ref new FileOpenPicker();
    fileOpenPicker->SuggestedStartLocation = PickerLocationId::MusicLibrary;
    fileOpenPicker->FileTypeFilter->Append(".wma");
    fileOpenPicker->FileTypeFilter->Append(".mp3");
    fileOpenPicker->FileTypeFilter->Append(".wav");

    // Get a music file from the picker
    task<StorageFile^> pickSingleFileTask = create_task(fileOpenPicker->PickSingleFileAsync());
    pickSingleFileTask.then([this](StorageFile^ storageFile)
    {
        if (storageFile == nullptr)
            return;

        // Disable button to avoid reentrancy problems & show progress bar
        openButton->IsEnabled = false;
        fileLoadProgressBar->Visibility = Windows::UI::Xaml::Visibility::Visible;
        fileLoadProgressBar->Value = 0;

        artistText->Text = "";
        albumText->Text = "";
        trackText->Text = "";
        durationText->Text = "";

        // Get the thumbnail for the display 
        task<StorageItemThumbnail^> getThumbnailTask = create_task(storageFile->GetThumbnailAsync(ThumbnailMode::SingleItem));
        getThumbnailTask.then([this](StorageItemThumbnail^ storageItemThumbnail)
        {
            task<void> setSourceTask(bitmapImage->SetSourceAsync(storageItemThumbnail));
            setSourceTask;
        });

        // Get information about the track
        StorageItemContentProperties^ contentProperties = storageFile->Properties;
        task<MusicProperties^> getMusicPropertiesTask = create_task(contentProperties->GetMusicPropertiesAsync());
        getMusicPropertiesTask.then([this](MusicProperties^ musicProperties)
        {
            artistText->Text = musicProperties->Artist;
            albumText->Text = musicProperties->Album;
            trackText->Text = musicProperties->Title;

            int durationInSeconds = int(musicProperties->Duration.Duration / 10000000);
            wchar_t str[16];
            swprintf(str, 16, L"%i:%02i", durationInSeconds / 60, 
                                          durationInSeconds % 60);
            durationText->Text = ref new String(str);
        });

        // Open the file for reading
        task<IRandomAccessStream^> openReadTask = create_task(storageFile->OpenAsync(FileAccessMode::Read));
        openReadTask.then([this](IRandomAccessStream^ randomAccessStream)
        {
            // Get rid of existing AudioFilePlayer
            if (pAudioFilePlayer != nullptr)
            {
                delete pAudioFilePlayer;
                pAudioFilePlayer = nullptr;
            }

            // Set record player to initial conditions
            recordRotate->Angle = 0;
            needleTransform->TranslateX = MAX_GROOVE_RADIUS;

            // Load the entire gigantic file into memory 
            task<LoadedAudioFileInfo> loadFileTask = create_task(LoadEntireFileAsync(randomAccessStream, &MainPage::FileLoadProgress));
            loadFileTask.then([this](LoadedAudioFileInfo loadedAudioFileInfo)
            {
                // Check for two channels of audio
                if (loadedAudioFileInfo.waveFormat.nChannels != 2)
                {
                    MessageDialog^ messageDialog = ref new MessageDialog("DeeJay can only handle files that have two channels of audio");
                    task<IUICommand^> showMessageDialogTask(messageDialog->ShowAsync());
                    showMessageDialogTask.then([](IUICommand^ uiCommand) { });
                }
                else
                {
                    // Draw spiral groove based on the length of the file
                    double durationInSeconds = double(loadedAudioFileInfo.bufferLength) / 
                                                      loadedAudioFileInfo.waveFormat.nSamplesPerSec;
                    int completeRevs = int(ceil(5 * durationInSeconds  / 9));
                    maxDurationInSeconds = 9 * completeRevs / 5.0;
                    float pixelsPerRev = (MAX_GROOVE_RADIUS - MIN_GROOVE_RADIUS) / float(completeRevs);
                    ConstructBezierSpiral(float(MAX_GROOVE_RADIUS), completeRevs, pixelsPerRev);

                    // Create an AudioFilePlayer and start it going
                    pAudioFilePlayer = new AudioFilePlayer(pXAudio2.Get(), loadedAudioFileInfo); 
                    pAudioFilePlayer->SetSpeedRatio(float(speedRatioSlider->Value));
                }

                // All done: re-enable button & hide progress bar
                openButton->IsEnabled = true;
                fileLoadProgressBar->Visibility = Windows::UI::Xaml::Visibility::Collapsed;
                }, task_continuation_context::use_current());
        });
    });
}

// Use a Bezier approximation to draw the spiral groove
void MainPage::ConstructBezierSpiral(float startRadius, int revolutions, float pixelsPerRev)
{
    spiralPathFigure->StartPoint = Point(startRadius, 0);
    spiralBezierSegment->Points->Clear();

    for (int rev = 0; rev < revolutions; rev++)
    {
        ConstructSingleBezierSpiralRotation(spiralBezierSegment->Points, startRadius, pixelsPerRev);
        startRadius -= pixelsPerRev;
    }
}

// Not quite correct: Code assumes tangents are to circle rather than spiral
void MainPage::ConstructSingleBezierSpiralRotation(PointCollection^ points, float radius, float pixelsPerRev)
{
    points->Append(Point(radius, -0.55f * radius));
    radius -= pixelsPerRev / 4;
    points->Append(Point(0.55f * radius, -radius));
    points->Append(Point(0, -radius));
    points->Append(Point(-0.55f * radius, -radius));
    radius -= pixelsPerRev / 4;
    points->Append(Point(-radius, -0.55f * radius));
    points->Append(Point(-radius, 0));
    points->Append(Point(-radius, 0.55f * radius));
    radius -= pixelsPerRev / 4;
    points->Append(Point(-0.55f * radius, radius));
    points->Append(Point(0, radius));
    points->Append(Point(0.55f * radius, radius));
    radius -= pixelsPerRev / 4;
    points->Append(Point(radius, 0.55f * radius));
    points->Append(Point(radius, 0));
}

// Touch input event handlers
void MainPage::OnRecordEllipseTapped(Object^ sender, TappedRoutedEventArgs^ args)
{
    if (pAudioFilePlayer != nullptr)
    {
        // Find distance from center
        Point point = args->GetPosition(recordEllipse);
        double distance = sqrt(pow(point.X - recordEllipse->ActualWidth / 2, 2) +
                               pow(point.Y - recordEllipse->ActualHeight / 2, 2));

        // Convert distance from 0 to 1
        double fraction = (MAX_GROOVE_RADIUS - distance) / (MAX_GROOVE_RADIUS - MIN_GROOVE_RADIUS);

        // Keep within bounds of 0 and 1
        fraction = max(0, min(1, fraction));

        pAudioFilePlayer->SetPositionFraction(fraction);
    }
}

void MainPage::OnRecordEllipseManipulationStarting(Object^ sender, ManipulationStartingRoutedEventArgs^ args)
{
    // Initialize for one-finger rotation
    Point point(float(recordEllipse->ActualWidth / 2), float(recordEllipse->ActualHeight / 2));
    args->Pivot = ref new ManipulationPivot(point, 100);
    smoothVelocity = 0;
}

void MainPage::OnRecordEllipseManipulationDelta(Object^ sender, ManipulationDeltaRoutedEventArgs^ args)
{
    if (pAudioFilePlayer != nullptr)
    {
        // Smooth out the velocities -- otherwise quite erratic
        smoothVelocity = 0.95 * smoothVelocity + 0.05 * args->Velocities.Angular / 0.2;
        pAudioFilePlayer->SetSpeedRatio(smoothVelocity);
    }
}

void MainPage::OnRecordEllipseManipulationCompleted(Object^ sender, ManipulationCompletedRoutedEventArgs^ args)
{
    if (pAudioFilePlayer != nullptr)
    {
        // Go back to what the slider says
        pAudioFilePlayer->SetSpeedRatio(float(speedRatioSlider->Value));
    }
}

// Asynchronous method to load file
task<LoadedAudioFileInfo> MainPage::LoadEntireFileAsync(IRandomAccessStream^ randomAccessStream,  ProgressReport progressReporter)
{
    return create_task([this, randomAccessStream, progressReporter] () -> LoadedAudioFileInfo
    {
        return LoadEntireFile(randomAccessStream, progressReporter);
    });
}

// Synchronous method to load file
LoadedAudioFileInfo MainPage::LoadEntireFile(IRandomAccessStream^ randomAccessStream, ProgressReport progressReporter)
{
    // The structure to return
    LoadedAudioFileInfo loadedAudioFileInfo = { 0 };

    // Create a IMFByteStream to wrap the IRandomAccessStream
    ComPtr<IMFByteStream> mfByteStream;
    HRESULT hresult = MFCreateMFByteStreamOnStreamEx((IUnknown *)randomAccessStream, &mfByteStream);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateMFByteStreamOnStreamEx failure");

    // Create the IMFSourceReader
    ComPtr<IMFSourceReader> mfSourceReader;
    hresult = MFCreateSourceReaderFromByteStream(mfByteStream.Get(), nullptr, &mfSourceReader);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateSourceReaderFromByteStream failure");

    // Create an IMFMediaType for setting the desired format
    ComPtr<IMFMediaType> mfMediaType;
    hresult = MFCreateMediaType(&mfMediaType);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateMediaType failure");

    hresult = mfMediaType->SetGUID(MF_MT_MAJOR_TYPE, MFMediaType_Audio);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFMediaType->SetGUID failure");

    hresult = mfMediaType->SetGUID(MF_MT_SUBTYPE, MFAudioFormat_PCM);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFMediaType->SetGUID failure");

    // Set the media type in the source reader
    hresult = mfSourceReader->SetCurrentMediaType(MF_SOURCE_READER_FIRST_AUDIO_STREAM, 
                                                  0, mfMediaType.Get());
    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFSourceReader->SetCurrentMediaType failure");

    // Get the resultant media type
    hresult = mfSourceReader->GetCurrentMediaType(MF_SOURCE_READER_FIRST_AUDIO_STREAM, &mfMediaType);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFSourceReader->GetCurrentMediaType failure");

    // Create a WAVEFORMATEX from the media type
    WAVEFORMATEX* pMediaWaveFormat;
    unsigned int waveFormatLength;
    hresult = MFCreateWaveFormatExFromMFMediaType(mfMediaType.Get(), &pMediaWaveFormat, &waveFormatLength);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "MFCreateWaveFormatExFromMFMediaType failure");

    // Copy it to the return structure and free the memory
    CopyMemory(&loadedAudioFileInfo.waveFormat, pMediaWaveFormat, sizeof(WAVEFORMATEX));
    CoTaskMemFree(pMediaWaveFormat);

    // Find the total duration
    PROPVARIANT propVar;
    hresult = mfSourceReader->GetPresentationAttribute(MF_SOURCE_READER_MEDIASOURCE, MF_PD_DURATION, &propVar);
    LONGLONG duration = propVar.uhVal.QuadPart;

    // Divide by 100nsec/sec and add an extra 3 seconds for safety's sake
    double durationInSeconds = duration / 10000000.0 + 3;

    // Allocate a buffer for the whole file
    int bufferLength = int(durationInSeconds * loadedAudioFileInfo.waveFormat.nAvgBytesPerSec);
    byte* audioBuffer = new byte[bufferLength];

    // Fill up the audio buffer
    int index = 0;

    while (true)
    {
        // Get an IMFSample object
        ComPtr<IMFSample> mfSample;
        DWORD flags = 0;
        HRESULT hresult = mfSourceReader->ReadSample(MF_SOURCE_READER_FIRST_AUDIO_STREAM, 
                                                     0, nullptr, &flags, nullptr, &mfSample);
        if (FAILED(hresult))
            throw ref new COMException(hresult, "MFSourceReader->ReadSample failure");

        // Check if we're at the end of the file
        if (flags & MF_SOURCE_READERF_ENDOFSTREAM)
        {
            break;
        }

        // If not, convert the data to a contiguous buffer
        ComPtr<IMFMediaBuffer> mfMediaBuffer;
        hresult = mfSample->ConvertToContiguousBuffer(&mfMediaBuffer);

        if (FAILED(hresult))
            throw ref new COMException(hresult, "IMFSample->ConvertToContiguousBuffer faiture");

        // Lock the audio buffer and copy the samples to local memory
        uint8 * pAudioData = nullptr;
        DWORD audioDataLength = 0;
        hresult = mfMediaBuffer->Lock(&pAudioData, nullptr, &audioDataLength); 

        if (FAILED(hresult))
            throw ref new COMException(hresult, "IMFMediaBuffer->Lock failure");

        // Let's hope the buffer is always large enough, but try to avoid crashing
        if (index + audioDataLength < unsigned int(bufferLength))
        {
            CopyMemory(audioBuffer + index, pAudioData, audioDataLength);
        }

        hresult = mfMediaBuffer->Unlock();

        if (FAILED(hresult))
            throw ref new COMException(hresult, "IMFMediaBuffer->Unlock");

        index += audioDataLength;

        // Report progress
        (this->*progressReporter)(double(index) / bufferLength);
    }
        
    // All done: Convert to AudioSample pointer
    loadedAudioFileInfo.bufferLength = index / sizeof(AudioSample);
    loadedAudioFileInfo.pBuffer = (AudioSample*) audioBuffer;

    return loadedAudioFileInfo;
}

// Report loading progress
void MainPage::FileLoadProgress(double progress)
{
    this->Dispatcher->RunAsync(CoreDispatcherPriority::Normal, ref new DispatchedHandler(
        [this, progress]()
        {
            fileLoadProgressBar->Value = 100 * progress;
        }));
}
