﻿//
// MainPage.xaml.h
// Declaration of the MainPage class.
//

#pragma once

#include <ppltasks.h>

#include "MainPage.g.h"
#include "AudioSample.h"
#include "LoadedAudioFileInfo.h"
#include "AudioFilePlayer.h"

namespace DeeJay
{
    public ref class MainPage sealed
    {
    private:
        static const int MAX_GROOVE_RADIUS = 480;
        static const int MIN_GROOVE_RADIUS = 220;
        typedef void (MainPage::*ProgressReport)(double);

        Microsoft::WRL::ComPtr<IXAudio2> pXAudio2;
        IXAudio2MasteringVoice * pMasteringVoice;
        AudioFilePlayer * pAudioFilePlayer;
        double maxDurationInSeconds;
        double smoothVelocity;

	public:
		MainPage();
	
    private:
        void OnSizeChanged(Platform::Object^ sender, Windows::UI::Xaml::SizeChangedEventArgs^ args);
        void OnCompositionTargetRendering(Platform::Object^ sender, Platform::Object^ args);

        void OnSpeedRatioSliderValueChanged(Platform::Object^ sender, Windows::UI::Xaml::Controls::Primitives::RangeBaseValueChangedEventArgs^ args);
        void OnVolumeSliderValueChanged(Platform::Object^ sender, Windows::UI::Xaml::Controls::Primitives::RangeBaseValueChangedEventArgs^ args);
        void OnOpenButtonClick(Platform::Object^ sender, Windows::UI::Xaml::RoutedEventArgs^ args);

        void ConstructBezierSpiral(float startRadius, int revolutions, float pixelsPerRev);
        void ConstructSingleBezierSpiralRotation(Windows::UI::Xaml::Media::PointCollection^ points, float radius, float pixelsPerRev);

        void OnRecordEllipseTapped(Platform::Object^ sender, Windows::UI::Xaml::Input::TappedRoutedEventArgs^ args);
        void OnRecordEllipseManipulationDelta(Platform::Object^ sender, Windows::UI::Xaml::Input::ManipulationDeltaRoutedEventArgs^ args);
        void OnRecordEllipseManipulationStarting(Platform::Object^ sender, Windows::UI::Xaml::Input::ManipulationStartingRoutedEventArgs^ args);
        void OnRecordEllipseManipulationCompleted(Platform::Object^ sender, Windows::UI::Xaml::Input::ManipulationCompletedRoutedEventArgs^ args);

        concurrency::task<LoadedAudioFileInfo> LoadEntireFileAsync(Windows::Storage::Streams::IRandomAccessStream^ randomAccessStream, ProgressReport progressReporter);
        LoadedAudioFileInfo LoadEntireFile(Windows::Storage::Streams::IRandomAccessStream^ randomAccessStream, ProgressReport progressReporter);
        void FileLoadProgress(double progress);
    };
}
